#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Implementation of the shared Exchange message bus and related classes for sending
and acting upon messages
"""

import asyncio
from collections import deque
from concurrent.futures import Future, ThreadPoolExecutor
import logging
import multiprocessing as mp
import os
from queue import Queue
from threading import get_ident, Event, Thread
import time
import traceback
from typing import Awaitable, Callable, NamedTuple, Sequence

import aiohttp

from . import eventloop

LOGGER = logging.getLogger(__name__)


_MESSAGE_FIELDS = {}

def format_type_name(ctype):
    """
    Convert a type or list of types to a string
    """
    if isinstance(ctype, Sequence):
        return '[{}]'.format(', '.join(map(format_type_name, ctype)))
    elif ctype is None:
        return 'None'
    return ctype.__name__

class ExchangeMessage:
    """
    A common base class for exchange messages
    """
    __slots__ = ('_values',)
    _fields = ()

    def __init__(self, *args, **kwargs):
        names, types, defaults, _positions = self._field_specs
        vals = []
        idx = 0
        if len(args) + len(kwargs) > len(names):
            raise TypeError("Too many arguments to constructor")
        for idx, name in enumerate(names):
            ftype = types.get(name)
            if idx < len(args):
                val = args[idx]
            else:
                if name in kwargs:
                    val = kwargs[name]
                elif name in defaults:
                    val = defaults[name]
                else:
                    raise TypeError("Property not provided to constructor: {}".format(name))
            if val is not None and ftype is not None and not isinstance(val, ftype):
                raise TypeError("Incorrect type for property '{}' ({}), expected {}".format(
                    name, format_type_name(type(val)), format_type_name(ftype)))
            vals.append(val)
        self._values = tuple(vals)

    @property
    def _field_specs(self):
        cname = self.__class__.__name__
        if cname not in _MESSAGE_FIELDS:
            names = []
            defaults = {}
            positions = {}
            types = {}
            for idx, field in enumerate(self._fields):
                if isinstance(field, tuple):
                    name = field[0]
                    if len(field) > 1:
                        types[name] = field[1]
                        if len(field) > 2:
                            defaults[name] = field[2]
                else:
                    name = field
                names.append(name)
                positions[name] = idx
            _MESSAGE_FIELDS[cname] = (names, types, defaults, positions)
        return _MESSAGE_FIELDS[cname]

    @property
    def _field_names(self):
        return self._field_specs[0]

    @property
    def _field_types(self):
        return self._field_specs[1]

    @property
    def _field_defaults(self):
        return self._field_specs[2]

    @property
    def _field_positions(self):
        return self._field_specs[3]

    def __iter__(self):
        return ((fname, self[idx]) for (idx, fname) in enumerate(self._field_names))

    def __getattr__(self, name):
        if name in self._field_names:
            return self._values[self._field_positions[name]]
        raise AttributeError("Unknown attribute: {}".format(name))

    def __getitem__(self, key):
        if isinstance(key, (slice, int)):
            return self._values[key]
        return getattr(self, key)

    def get(self, name: str, defval=None):
        """
        Get a property of the message by name

        Args:
            name: the property name
            defval: the default value to return if the property is not defined
        """
        return getattr(self, name, defval)

    def __repr__(self):
        cls = self.__class__.__name__
        params = ['{}={}'.format(fname, self[idx]) for (idx, fname) in enumerate(self._field_names)]
        return '{}({})'.format(cls, ', '.join(params))


class ExchangeFail(ExchangeMessage):
    """
    An error class to represent an exception in message processing

    This is not a subclass of :class:`Exception` as that cannot be pickled
    and transported over the message bus
    """
    _fields = ('value', 'exc_info')
    def __init__(self, value, exc_info=True):
        if exc_info is True:
            # cannot pass real exception or traceback through the message pipe
            exc_info = traceback.format_exc()
        super(ExchangeFail, self).__init__(value, exc_info)

    def format(self) -> str:
        """
        Format this :class:`ExchangeFail` instance as a string including the
        traceback, if any
        """
        ret = '{}'.format(self.value)
        if self.exc_info:
            ret += "\n" + str(self.exc_info)
        return ret

    def __repr__(self):
        cls = self.__class__.__name__
        ret = '{}(value={})'.format(cls, self.value)
        if self.exc_info:
            ret += "\n" + str(self.exc_info)
        return ret


class StopMessage(ExchangeMessage):
    """
    Basic stop-processing message for :class:`MessageProcessor` instances
    """
    pass


MessageWrapper = NamedTuple('MessageWrapper', [
    ('from_pid', str),
    ('ident', str),
    ('message', ExchangeMessage),
    ('ref', str)])
MessageWrapper.__new__.__defaults__ = (None,)
MessageWrapper.__doc__ = """
    A wrapper for a message being passed through the :class:`Exchange` message bus

    Attributes:
        from_pid (str): The identifier of the sending service
        ident (str): A unique identifier for the message, used to tag responses
        message (ExchangeMessage): The message received
        ref (str): An optional identifier for the message being responded to
    """

QueuedMessage = NamedTuple('QueuedMessage', [
    ('to_pid', str),
    ('message', ExchangeMessage)])
QueuedMessage.__doc__ = """
    A wrapper for a message queued to be sent to the exchange

    Attributes:
        to_pid (str): The identifier of the recipient service
        message (ExchangeMessage): The message to be sent
    """


class Exchange:
    """
    A central message exchange hub for receiving requests and passing them to processors
    which may live in a different thread or process, but have a known identifier.
    Multiple processors may also respond to the same identifier in order to share processing.
    Responses are optional and can be tied to the original request.
    """

    def __init__(self):
        self._cmd_pipe = mp.Pipe()
        self._cmd_lock = mp.Lock()
        self._proc = None
        self._req_cond = mp.Condition(mp.Lock())

    def start(self, process: bool = True) -> None:
        """
        Start the message exchange as a thread or process
        """
        if process:
            evt = mp.Event()
            proc = mp.Process(target=self._run, args=(evt,))
        else:
            evt = Event()
            proc = Thread(target=self._run, args=(evt,))
        proc.daemon = True
        proc.start()
        evt.wait()
        self._proc = proc
        LOGGER.info('Started exchange')

    def stop(self, drain: bool = True) -> None:
        """
        Send a stop signal to the polling thread
        """
        LOGGER.info('Stopping exchange')
        with self._req_cond:
            self._cmd('stop', drain)
            # wake all threads waiting for an incoming message
            self._req_cond.notify_all()

    def join(self) -> None:
        """
        Wait for the exchange to finish running
        """
        self._proc.join()

    def status(self) -> dict:
        """
        Retrieve the status from the polling thread

        Returns:
            A dict in the form {'pending': int, 'processed': int, 'total': int}
            representing the total numbers of messages handled by the exchange
        """
        with self._req_cond:
            return self._cmd('status')

    def _cmd(self, *command):
        """
        Execute a command against the exchange, using a process lock to synchronize
        requests and responses.
        Supported commands are currently `send`, `recv`, `status` and `stop`
        """
        with self._cmd_lock:
            self._cmd_pipe[1].send(command)
            return self._cmd_pipe[1].recv()

    def register(self, to_pid: str) -> bool:
        """
        Register a listener on the exchange
        """
        return self._cmd('register', to_pid)

    def is_registered(self, to_pid: str) -> bool:
        """
        Check if a listener is currently running
        """
        return self._cmd('check', to_pid)

    def send(self, to_pid: str, wrapper: MessageWrapper) -> bool:
        """
        Add a message to the bus, blocking until the processing thread is ready

        Args:
            to_pid: The identifier for the receiving service
            wrapper: The message to be added to the queue

        Returns:
            True if the message is successfully added to the queue
        """
        # Blocks until we have access to the message queues and command pipe
        # FIXME add a maximum buffer size for the message queues and allow blocking
        # until there is room in the buffer (optional blocking=True argument)
        with self._req_cond:
            LOGGER.debug('send to %s/%s %s', to_pid, wrapper.ref, wrapper.message)
            status = self._cmd('send', to_pid, wrapper)
            # wake all threads waiting for an incoming message
            self._req_cond.notify_all()
        return status

    def recv(self, to_pid: str, blocking: bool = True, timeout=None) -> MessageWrapper:
        """
        Receive a message from the bus

        Args:
            to_pid: The identifier of the recipient service
            blocking: Whether to sleep this thread until a message is received
            timeout: An optional timeout before aborting

        Returns:
            The next message in the queue, or None
        """
        #pylint: disable=broad-except
        try:
            LOGGER.debug('recv %s', to_pid)
            locked = self._req_cond.acquire(blocking)
            message = None
            if locked:
                message = self._cmd('recv', to_pid)
                while message is None and (blocking or timeout is not None):
                    #LOGGER.warning("Locked (%r), wait", locked)
                    locked = self._req_cond.wait(timeout)
                    if locked:
                        message = self._cmd('recv', to_pid)
                    if not locked or timeout is not None:
                        break
                if locked:
                    self._req_cond.release()
        except Exception:
            LOGGER.exception('Error in recv:')
            raise
        return message

    def _drain(self) -> None:
        while self._cmd('drain'):
            time.sleep(1)

    def _run(self, event: Event) -> None:
        """
        The message processing loop
        """
        drain = Thread(target=self._drain)
        drain.start()
        #pylint: disable=broad-except
        pending = 0
        processed = {}
        queue = {}
        stop_time = None
        event.set()
        try:
            while True:
                command = self._cmd_pipe[0].recv()
                if command[0] == 'register':
                    to_pid = command[1]
                    if to_pid and to_pid not in queue:
                        queue[to_pid] = deque()
                        self._cmd_pipe[0].send(True)
                        LOGGER.debug("registered %s", to_pid)
                    else:
                        self._cmd_pipe[0].send(False)
                elif command[0] == 'check':
                    to_pid = command[1]
                    self._cmd_pipe[0].send(to_pid and to_pid in queue)
                elif command[0] == 'send':
                    if stop_time:
                        LOGGER.debug("rejected message %s %s", command[1], command[2])
                        self._cmd_pipe[0].send(False)
                    else:
                        to_pid = command[1]
                        if to_pid in queue:
                            queue[to_pid].append(command[2])
                            pending += 1
                        else:
                            self._cmd_pipe[0].send(True)
                        self._cmd_pipe[0].send(True)
                elif command[0] == 'recv':
                    to_pid = command[1]
                    wrapper = None
                    if to_pid in queue:
                        try:
                            wrapper = queue[to_pid].popleft()
                            processed[to_pid] = processed.get(to_pid, 0) + 1
                            pending -= 1
                        except IndexError:
                            pass
                        if wrapper and isinstance(wrapper.message, StopMessage):
                            pending -= len(queue[to_pid])
                            del queue[to_pid]
                            LOGGER.debug("unregistered %s", to_pid)
                    self._cmd_pipe[0].send(wrapper)
                elif command[0] == 'status':
                    total = sum(processed.values())
                    self._cmd_pipe[0].send({
                        'pending': pending,
                        'processed': processed,
                        'total': total})
                elif command[0] == 'drain':
                    # clean up expired messages ...
                    if stop_time:
                        if not pending or time.time() - stop_time >= 5:
                            if pending:
                                LOGGER.debug("terminating with %s messages pending", pending)
                            self._cmd_pipe[0].send(False)
                            break
                    self._cmd_pipe[0].send(True)
                elif command[0] == 'stop':
                    for to_pid in queue:
                        LOGGER.debug("ordering %s to stop", to_pid)
                        queue[to_pid].append(MessageWrapper(None, None, StopMessage()))
                        pending += 1
                    stop_time = time.time()
                    self._cmd_pipe[0].send(True)
                else:
                    raise ValueError('Unrecognized command: {}'.format(command[0]))
        except Exception:
            LOGGER.exception('Error in exchange:')


class MessageTarget:
    """
    A wrapper for sending messages to a single target.

    Example:
        >>> target = MessageTarget(target_pid, exchange, my_pid)
        >>> target.send_noreply('hello')
        True
    """

    def __init__(self, pid: str, exchange: Exchange, from_pid: str = None):
        self._pid = pid
        self._from_pid = from_pid
        self._exchange = exchange

    @property
    def pid(self) -> str:
        """
        Accessor for the identifier of the recipient service
        """
        return self._pid

    @property
    def exchange(self) -> Exchange:
        """
        Accessor for the :class:`Exchange` used by this target
        """
        return self._exchange

    @property
    def from_pid(self) -> str:
        """
        Accessor for the identifier of the sending service
        """
        return self._from_pid

    def _send_message(self, message: MessageWrapper) -> bool:
        """
        Perform the actual addition to the message queue

        Args:
            message: the message to be sent
        """
        return self._exchange.send(self._pid, message)

    def send(
            self,
            ident: str,
            message: ExchangeMessage,
            ref: str = None,
            from_pid: str = None) -> bool:
        """
        Send a message to the recipient service

        Args:
            ident: The identifier used by the message response
            message: The message being sent
            ref: An optional identifier for the message being responded to
            from_pid: An optional override for the sender identifier

        Returns:
            True if the message was successfully added to the queue
        """
        return self._send_message(MessageWrapper(
            from_pid if from_pid is not None else self._from_pid,
            ident,
            message,
            ref))

    def send_noreply(
            self,
            message: ExchangeMessage,
            ref: str = None,
            from_pid: str = None) -> bool:
        """
        Send a message with no reply expected

        Returns:
            True if the message was successfully added to the queue
        """
        return self.send(None, message, ref, from_pid)


class MessageProcessor:
    """
    A generic message processor which polls the exchange for messages sent to
    this endpoint and runs the abstract 'process' method to perform actions
    and send responses.
    """

    def __init__(self, pid: str, exchange: Exchange):
        self._pid = pid
        self._exchange = exchange
        self._poll_thread = None

    @property
    def pid(self) -> str:
        """
        Accessor for the identifier of this request processor service
        """
        return self._pid

    @property
    def exchange(self) -> Exchange:
        """
        Accessor for the :class:`Exchange` used by this request processor
        """
        return self._exchange

    def get_message_target(self, pid: str) -> MessageTarget:
        """
        Quickly create a :class:`MessageTarget` for a service on the same message bus
        """
        return MessageTarget(pid, self._exchange, self._pid)

    def start(self, _wait: bool = True) -> None:
        """
        Run a thread to poll for received messages
        """
        self._poll_thread = Thread(target=self._run)
        self._poll_thread.start()

    def _start_run(self) -> bool:
        """
        Perform any additional initializion in polling thread
        """
        return self._exchange.register(self._pid)

    def join(self) -> None:
        """
        Await our polling thread. `stop()` must be called in order to cause it to abort
        """
        if self._poll_thread:
            self._poll_thread.join()

    def send_stop_message(self) -> bool:
        """
        Send the service a stop signal to end processing
        """
        return self.send_noreply(self._pid, StopMessage())

    def stop(self, wait: bool = True) -> None:
        """
        Send a stop signal to the polling thread in order to abort polling
        """
        if self.send_stop_message():
            while wait and self._exchange.is_registered(self._pid):
                time.sleep(0.01)

    def _stop_run(self) -> None:
        """
        Perform any additional shutdown actions in polling thread
        """
        pass

    def _run(self) -> None:
        """
        The main thread run loop
        """
        if not self._start_run():
            return
        self._poll_messages()
        self._stop_run()

    def _poll_messages(self) -> None:
        """
        The polling loop for receiving messages from the exchange
        """
        #pylint: disable=broad-except
        try:
            while self._poll_message():
                pass
        except Exception:
            LOGGER.exception('Exception while processing messages:')

    def _poll_message(self) -> bool:
        """
        Wait for a message from the exchange
        """
        #pylint: disable=broad-except
        # blocks until a message is available
        received = self._exchange.recv(self._pid)
        LOGGER.debug('%s processing message: %s', self._pid, received.message)
        if isinstance(received.message, StopMessage):
            return False
        try:
            if self._process_message(received) is False:
                return False
        except Exception:
            LOGGER.exception('Exception during message processing:')
            errmsg = ExchangeFail('Exception during message processing', True)
            self._reply_with_error(received, errmsg)
        return True

    def _reply_with_error(
            self,
            from_message: MessageWrapper,
            errmsg: ExchangeFail) -> bool:
        """
        Send an error message back to the sender of a previous message

        Args:
            from_message: the message which triggered the error
            errmsg: the error message to be sent
        """
        if isinstance(from_message.message, ExchangeFail):
            LOGGER.error(from_message.message.format())
            return False
        return self.send_noreply(from_message.from_pid, errmsg, from_message.ident)

    def _send_message(self, to_pid: str, wrapper: MessageWrapper) -> bool:
        """
        Perform the actual addition to the exchange message queue

        Args:
            to_pid: the identifier of the recipient
            message: the message to be sent
        """
        return self._exchange.send(to_pid, wrapper)

    def send(
            self,
            to_pid: str,
            ident: str,
            message: ExchangeMessage,
            ref: str = None,
            from_pid: str = None) -> bool:
        """
        Send a message to a recipient on the exchange

        Args:
            to_pid: The identifier of the recipient
            ident: The identifier of thie message, to be used by responses
            message: The content of the message
            ref: The identifier of the message being responded to
            from_pid: An optional override for the sender identifier

        Returns:
            True if the message was successfully added to the queue
        """
        return self._send_message(
            to_pid,
            MessageWrapper(from_pid or self._pid, ident, message, ref))

    def send_noreply(
            self,
            to_pid: str,
            message: ExchangeMessage,
            ref: str = None,
            from_pid: str = None) -> bool:
        """
        Send a message with no reply expected

        Returns:
            True if the message was successfully added to the queue
        """
        return self._send_message(
            to_pid,
            MessageWrapper(from_pid or self._pid, None, message, ref))

    def _process_message(self, received: MessageWrapper) -> bool:
        """
        Process a message from another service and optionally send a message in response

        Returns: `False` if the polling thread should terminate
        """
        pass


class LoggingTCPConnector(aiohttp.TCPConnector):
    def _release(self, key, protocol, *, should_close=False):
        close = should_close or self._force_close or protocol.should_close
        LOGGER.debug("Connection released: %s", close and "Closing" or "Leaving open")
        super(LoggingTCPConnector, self)._release(key, protocol, should_close=should_close)


class RequestExecutor(MessageProcessor):
    """
    An subclass of :class:`MessageProcessor` which starts a thread for each outgoing request
    to wait for responses. One of these should live in each process which wants to perform
    async requests via the :class:`Exchange` (like a webserver process). It normally assumes
    that all incoming messages are simply responses to earlier requests.
    Processing should not block the main thread (much) to avoid breaking asyncio.
    """

    def __init__(self, pid: str, exchange: Exchange):
        super(RequestExecutor, self).__init__(pid, exchange)
        self._connector = None
        self._out_queue = None
        self._req_lock = None
        self._requests = {}
        self._runner = None

    def start(self, wait: bool = True) -> None:
        """
        Initialize our :class:`eventloop.Runner` and run our polling thread to listen for messages
        """
        self._out_queue = Queue()
        self._runner = eventloop.Runner()
        self._runner.start(wait)
        self._req_lock = asyncio.Lock(loop=self._runner.loop)
        # Poll for results in a thread from our thread pool
        self.run_thread(self._run, ident='polling thread {}'.format(self.pid))

    def _start_run(self) -> bool:
        if not super(RequestExecutor, self)._start_run():
            return False
        # Send outgoing messages to the exchange (without blocking our event loop)
        self.run_thread(self._send_messages, ident='sending thread {}'.format(self.pid))
        return True

    # In the webserver environment, the process we're concerned with has already started
    # so just use start() instead
    def start_process(self) -> mp.Process:
        """
        Start this executor in a new process
        """
        def _start():
            self._init_process()
            self.start()
            self._runner.join()
        asyncio.get_child_watcher()
        proc = mp.Process(target=_start)
        proc.start()
        return proc

    def runner(self) -> eventloop.Runner:
        """
        Accessor for the event loop runner instance used to execute tasks
        """
        return self._runner

    def _stop_run(self) -> None:
        """
        Stop our sending thread and any other tasks in progress
        """
        # stop sending messages
        self._out_queue.put_nowait(None)
        self._out_queue.join()
        # close TCP connector
        if self._connector:
            self._connector.close()
        # shut down event loop
        self._runner.stop()

    def run_task(self, proc: Awaitable) -> asyncio.Future:
        """
        Add a coroutine task to be performed by the runner

        Args:
            proc: the coroutine to be executed in the runner's event loop
        """
        return self._runner.run_task(proc)

    def run_thread(self, proc: Callable, *args, ident: str = None) -> asyncio.Future:
        """
        Add a task to be processed, as either a coroutine or function

        Args:
            proc: the function to be run in the :class:`ThreadPoolExecutor`
            args: arguments to pass to the proc, if a function
        """
        if ident and False:
            _proc = proc
            def proc(*args):
                tid = get_ident()
                LOGGER.info(">> start thread %s %s", ident, tid)
                ret = _proc(*args)
                LOGGER.info("<< end thread %s %s", ident, tid)
                return ret
        return self._runner.run_in_executor(None, proc, *args)

    def _init_process(self) -> None:
        """
        Initialize ourselves in a newly started process
        """
        # create new event loop after fork
        asyncio.get_event_loop().close()
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    def _send_messages(self) -> None:
        """
        Thread loop for sending messages added to the out-queue
        """
        while True:
            msg = self._out_queue.get()
            if msg is None:
                self._out_queue.task_done()
                break
            self._exchange.send(msg.to_pid, msg.message)
            self._out_queue.task_done()

    def _send_message(self, to_pid: str, wrapper: MessageWrapper) -> bool:
        """
        Add the message to our out-queue for processing instead of sending directly

        Args:
            to_pid: the identifier of the recipient
            message: the message to be sent
        """
        self._out_queue.put_nowait(QueuedMessage(to_pid, wrapper))
        return True

    async def _send_request(self, to_pid: str, request: ExchangeMessage,
                            future: Future, timeout: int = None) -> None:
        """
        Send a request to a target service on the exchange and add it to our
        collection to automatically associate the response later

        Args:
            to_pid: the target service identifier
            request: the message payload
            future: used to return the response to (potentially) another thread
            timeout: an optional timeout before cancelling the request
        """
        message = MessageWrapper(self._pid, os.urandom(10), request)
        result = None
        async with self._req_lock:
            if message.ident in self._requests:
                future.set_exception(RuntimeError('Duplicate request identifier'))
                return
            self._requests[message.ident] = future
        result = self._send_message(to_pid, message)
        if not result:
            future.set_exception(RuntimeError('Request could not be processed'))
        elif timeout:
            self.run_task(self._cancel_request(message.ident, timeout))

    async def _cancel_request(self, ident: str, timeout: int = None) -> None:
        """
        Cancel an outstanding request

        Args:
            ident: the request identifier
            timeout: an optional timeout to wait before cancelling
        """
        if timeout:
            await asyncio.sleep(timeout)
        async with self._req_lock:
            if ident in self._requests and not self._requests[ident].done():
                self._requests[ident].cancel()

    def submit(
            self,
            to_pid: str,
            request: ExchangeMessage,
            timeout: int = None) -> asyncio.Future:
        """
        Submit a message to another service and run a task to poll for the results

        Args:
            to_pid: the identifier of the target service
            request: the body of the message to be sent
            timeout: an optional timeout to wait before cancelling the request
        """
        result = Future()
        self.run_task(self._send_request(to_pid, request, result, timeout))
        return asyncio.wrap_future(result)

    async def _handle_message(self, received: MessageWrapper) -> bool:
        """
        Handle a message received from another service on the exchange by awaking
        any tasks waiting for results

        Args:
            received: the received message to be processed
        """
        result = False
        if received.ref:
            async with self._req_lock:
                if received.ref in self._requests:
                    if not self._requests[received.ref].cancelled():
                        self._requests[received.ref].set_result(received.message)
                    result = True
                self._requests = {
                    ident: req for ident, req in self._requests.items() if not req.done()}
        return result

    async def _handle_message_task(self, received: MessageWrapper) -> None:
        """
        Handle message processing within our own event loop

        Args:
            received: the message received from the exchange
        """
        #pylint: disable=broad-except
        try:
            if not await self._handle_message(received):
                LOGGER.debug('unhandled message to %s/%s from %s: %s',
                             self._pid, received.ref, received.from_pid, received.message)
        except Exception:
            errmsg = ExchangeFail('Exception during message processing', True)
            self._reply_with_error(received, errmsg)

    def _process_message(self, received: MessageWrapper) -> bool:
        """
        Handle a message received from another service on the exchange

        Args:
            received: the received message to be processed
        """
        # push the handling of the message into our own event loop
        self.run_task(self._handle_message_task(received))
        return True

    @property
    def tcp_connector(self) -> aiohttp.TCPConnector:
        """
        Return a connection pool associated with this event loop, which allows HTTP
        connection reuse
        """
        if not self._connector:
            force_close = os.getenv('HTTP_FORCE_CLOSE_CONNECTIONS')
            force_close = bool(force_close) and force_close != 'false'
            self._connector = LoggingTCPConnector(force_close=force_close)
        return self._connector

    def http_client(self, *args, **kwargs) -> aiohttp.ClientSession:
        """
        Construct an HTTP client using the shared connection pool
        """
        no_reuse = os.getenv('HTTP_NO_CONNECTOR_REUSE')
        no_reuse = bool(no_reuse) and no_reuse != 'false'
        if 'connector' not in kwargs and not no_reuse:
            kwargs['connector'] = self.tcp_connector
            kwargs['connector_owner'] = False
        keep_cookies = os.getenv('HTTP_PRESERVE_COOKIES')
        keep_cookies = bool(keep_cookies) and keep_cookies != 'false'
        if 'cookie_jar' not in kwargs and not keep_cookies:
            kwargs['cookie_jar'] = aiohttp.DummyCookieJar()
        return aiohttp.ClientSession(*args, **kwargs)

    @property
    def http(self):
        """
        A quick accessor for a default HTTP client instance
        """
        return self.http_client()

    def get_request_target(self, pid: str) -> 'RequestTarget':
        """
        Create a :class:`RequestTarget` for a specific service

        Args:
            pid: the identifer of the target service
        """
        return RequestTarget(self, pid)


class RequestTarget:
    """
    An endpoint for a :class:`RequestExecutor` which uses submit() to poll
    for responses to requests. It must be created within the same process as the
    executor instance

    Example:
        >>> target = RequestTarget(executor, target_pid)
        >>> target.request('hello')
        Future<...>
    """

    def __init__(self, executor: RequestExecutor, pid: str):
        self._executor = executor
        self._pid = pid

    @property
    def pid(self):
        """
        Accessor for the target service identifier
        """
        return self._pid

    @property
    def executor(self):
        """
        Accessor for the :class:`RequestExecutor` instance
        """
        return self._executor

    def request(self, message: ExchangeMessage, timeout: int = None) -> asyncio.Future:
        """
        Send a request to the recipient service, awaiting the response in
        a method defined by the executor

        Args:
            message: The message to be sent
            timeout: An optional timeout for the message response
        """
        return self._executor.submit(
            self.pid,
            message,
            timeout)


class HelloProcessor(MessageProcessor):
    """
    A simple request processor for testing response functionality or stress testing
    """
    def _process_message(self, received: MessageWrapper) -> bool:
        self.send_noreply(received.from_pid,
                          'hello from {} {}'.format(os.getpid(), get_ident()), received.ident)


class ThreadedHelloProcessor(HelloProcessor):
    """
    A threaded request processor for testing delayed, blocking and non-blocking responses
    """
    def __init__(self, pid, exchange, blocking=False, max_workers=5):
        super(ThreadedHelloProcessor, self).__init__(pid, exchange)
        self._blocking = blocking
        self._pool = None
        self._max_workers = max_workers

    def start(self, _wait: bool = True) -> None:
        self._pool = ThreadPoolExecutor(self._max_workers) #thread_name_prefix=self._pid
        self._pool.submit(self._run)

    def start_process(self) -> mp.Process:
        """
        Start this demo processor as a process instead of a thread
        """
        proc = mp.Process(target=lambda: self.start().result())
        proc.start()
        return proc

    def _process_message(self, received: MessageWrapper) -> bool:
        if self._blocking:
            self._delayed_process(received)
        else:
            self._pool.submit(self._delayed_process, received)

    def _delayed_process(self, received: MessageWrapper) -> bool:
        time.sleep(1)
        return super(ThreadedHelloProcessor, self)._process_message(received)


# Testing two workers dividing requests:
# hello = ThreadedHelloProcessor('hello', exchange, blocking=True)
# hello.start_process()
# hello.start_process()
# .. exchange.send('hello', None, None, 'poke') ..



#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Implementation of the standard manager class for the :class:`IndyService`
"""

import logging
from typing import Mapping

from ..common.config import load_config
from ..common.manager import ConfigServiceManager
from .client import IndyClient
from .config import IndyConfigError, SchemaManager
from .service import IndyService
from .tob import CRED_TYPE_PARAMETERS

LOGGER = logging.getLogger(__name__)


def load_credential_type(ctype, schema_mgr: SchemaManager) -> dict:
    """
    Load the credential types defined by our config into a standard format
    """
    if "topic" not in ctype:
        raise IndyConfigError("Credential type must define 'topic'")
    if "schema" not in ctype:
        raise IndyConfigError("Credential type must define 'schema'")
    if isinstance(ctype["schema"], str):
        name = ctype["schema"]
        version = None
        origin_did = None
        attributes = None
    elif isinstance(ctype["schema"], dict):
        name = ctype["schema"].get("name")
        version = ctype["schema"].get("version")
        origin_did = ctype["schema"].get("origin_did")
        attributes = ctype["schema"].get("attributes")
    else:
        raise IndyConfigError("Credential type schema must be string or dict")
    if not name:
        raise IndyConfigError("Credential type schema missing 'name'")
    if not version or not (attributes or origin_did):
        schema = schema_mgr.find(name, version)
        if schema:
            version = schema.version
            attributes = schema.attr_names
            origin_did = schema.origin_did
        else:
            raise IndyConfigError(
                "Schema definition not found: {} {}".format(name, version)
            )
    ret = {
        "schema_name": name,
        "schema_version": version,
        "origin_did": origin_did,
        "attributes": attributes,
        "params": {},
    }
    for k in CRED_TYPE_PARAMETERS:
        if k in ctype:
            ret["params"][k] = ctype[k]
    return ret


class IndyManager(ConfigServiceManager):
    """
    A manager for initializing the Indy service from standard configuration files
    """

    def __init__(self, env: Mapping = None, pid: str = 'manager'):
        super(IndyManager, self).__init__(env, pid)
        self._schema_mgr = None

    def _init_services(self):
        """
        Initialize the Indy service
        """
        super(IndyManager, self)._init_services()

        indy = self.init_indy_service()
        self.add_service("indy", indy)

    def get_client(self) -> IndyClient:
        """
        Obtain an IndyClient attached to the registered Indy service
        """
        return IndyClient(self.get_service_request_target("indy"))

    def get_service_init_params(self) -> dict:
        """
        Get a dictionary of parameters for initializing the :class:`IndyService` instance
        """
        genesis_path = self._env.get("INDY_GENESIS_PATH")
        if not genesis_path:
            raise IndyConfigError(
                "Indy genesis transaction path (INDY_GENESIS_PATH) not defined"
            )
        genesis_url = self._env.get("INDY_GENESIS_URL")
        ledger_url = self._env.get("INDY_LEDGER_URL")
        if not genesis_url and not ledger_url:
            raise IndyConfigError("Neither INDY_GENESIS_URL nor INDY_LEDGER_URL are defined")
        protocol_version = self._env.get("LEDGER_PROTOCOL_VERSION") or None
        return {
            "auto_register": self._env.get("AUTO_REGISTER_DID", 1),
            "genesis_path": genesis_path,
            "ledger_url": ledger_url,
            "genesis_url": genesis_url,
            "protocol_version": protocol_version,
        }

    def init_indy_service(self, pid: str = "indy") -> IndyService:
        """
        Initialize the Hyperledger Indy service

        Args:
            pid: the identifier for the :class:`IndyService` instance
        """
        spec = self.get_service_init_params()
        LOGGER.info("Initializing Indy service")
        return IndyService(pid, self._exchange, self._env, spec)

    async def _service_start(self) -> bool:
        """
        After running processing load our schemas and initialize all services
        """
        ret = await super(IndyManager, self)._service_start()
        if ret:
            self._schema_mgr = self._load_schemas()
            self.run_task(self._load_config())
        return ret

    def _load_schemas(self) -> SchemaManager:
        """
        Load all configured schemas
        """
        mgr = SchemaManager()
        std = load_config('vonx.config:schemas.yml')
        if std:
            mgr.load(std)
        ext = self.load_config_path('SCHEMAS_CONFIG_PATH', 'schemas.yml')
        if ext:
            mgr.load(ext)
        return mgr

    async def _load_config(self) -> None:
        """
        Initialize our client and populate services based on configuration
        """
        client = self.get_client()
        await self._register_agents(client)
        await self._register_proof_requests(client)
        await client.sync(False)

    async def _register_agents(self, client: IndyClient) -> None:
        """
        Load agent settings from our configuration files
        """
        limit_agents = self._env.get("AGENTS")
        limit_agents = (
            set(limit_agents.split())
            if (limit_agents and limit_agents != "all")
            else None
        )

        await self._register_issuers(client, limit_agents)
        await self._register_holders(client, limit_agents)
        await self._register_verifiers(client, limit_agents)

    async def _register_issuers(self, client: IndyClient, limit_agents: set):
        """
        Register all issuer services from the configuration
        """
        issuers = []
        issuer_ids = []
        config_issuers = self.services_config("issuers")
        if not config_issuers:
            LOGGER.debug("No issuers defined by configuration")
        for issuer_key, issuer_cfg in config_issuers.items():
            if not issuer_cfg.get("id"):
                issuer_cfg["id"] = issuer_key
            if limit_agents is None or issuer_cfg["id"] in limit_agents:
                issuers.append(issuer_cfg)
                issuer_ids.append(issuer_cfg["id"])
        if issuers:
            for issuer_cfg in issuers:
                await self._register_issuer(client, issuer_cfg)
        elif config_issuers:
            LOGGER.warning("No defined issuers referenced by AGENTS")

    async def _register_issuer(self, client: IndyClient, issuer_cfg: dict) -> str:
        """
        Register a single issuer service from the configuration
        """
        issuer_id = issuer_cfg["id"]
        if "wallet" not in issuer_cfg:
            raise IndyConfigError("Wallet not defined for issuer: {}".format(issuer_id))
        wallet_cfg = issuer_cfg["wallet"]
        del issuer_cfg["wallet"]
        if "credential_types" not in issuer_cfg:
            raise IndyConfigError("Missing credential_types for issuer: {}".format(issuer_id))
        cred_types = issuer_cfg["credential_types"]
        del issuer_cfg["credential_types"]
        if "connection" not in issuer_cfg:
            raise IndyConfigError("Missing connection for issuer: {}".format(issuer_id))
        connection_cfg = issuer_cfg["connection"]
        del issuer_cfg["connection"]

        if not wallet_cfg.get("name"):
            wallet_cfg["name"] = issuer_id + "-Issuer-Wallet"
        if not wallet_cfg.get("seed"):
            raise IndyConfigError("Missing wallet seed for issuer: {}".format(issuer_id))
        wallet_id = await client.register_wallet(wallet_cfg)
        issuer_id = await client.register_issuer(wallet_id, issuer_cfg)

        for type_spec in cred_types:
            cred_type = load_credential_type(type_spec, self._schema_mgr)
            await client.register_credential_type(
                issuer_id,
                cred_type["schema_name"],
                cred_type["schema_version"],
                cred_type["origin_did"],
                cred_type["attributes"],
                cred_type["params"],
            )

        if connection_cfg:
            if not connection_cfg.get("id"):
                connection_cfg["id"] = issuer_id
            conn_type = connection_cfg.get("type", "TheOrgBook")
            if conn_type == "TheOrgBook":
                _conn_id = await client.register_orgbook_connection(
                    issuer_id, connection_cfg)
            else:
                _conn_id = await client.register_http_connection(
                    issuer_id, connection_cfg)
        return issuer_id

    async def _register_holders(self, client: IndyClient, limit_agents: set):
        """
        Register all holder services from the configuration
        """
        holders = []
        holder_ids = []
        config_holders = self.services_config("holders")
        if not config_holders:
            LOGGER.debug("No holders defined by configuration")
        for holder_key, holder_cfg in config_holders.items():
            if not holder_cfg.get("id"):
                holder_cfg["id"] = holder_key
            if limit_agents is None or holder_cfg["id"] in limit_agents:
                holders.append(holder_cfg)
                holder_ids.append(holder_cfg["id"])
        if holders:
            for holder_cfg in holders:
                await self._register_holder(client, holder_cfg)
        elif config_holders:
            LOGGER.info("No defined holders referenced by AGENTS")

    async def _register_holder(self, client: IndyClient, holder_cfg: dict) -> str:
        """
        Register a single holder service from the configuration
        """
        holder_id = holder_cfg["id"]
        if "wallet" not in holder_cfg:
            raise IndyConfigError("Wallet not defined for holder: {}".format(holder_id))
        wallet_cfg = holder_cfg["wallet"]
        del holder_cfg["wallet"]
        if not wallet_cfg.get("name"):
            wallet_cfg["name"] = holder_id + "-Holder-Wallet"
        if not wallet_cfg.get("seed"):
            raise IndyConfigError("Missing wallet seed for holder: {}".format(holder_id))
        wallet_id = await client.register_wallet(wallet_cfg)
        holder_id = await client.register_holder(wallet_id, holder_cfg)
        return holder_id

    async def _register_verifiers(self, client: IndyClient, limit_agents: set):
        """
        Register all verifier services from the configuration
        """
        verifiers = []
        verifier_ids = []
        config_verifiers = self.services_config("verifiers")
        if not config_verifiers:
            LOGGER.debug("No verifiers defined by configuration")
        for verifier_key, verifier_cfg in config_verifiers.items():
            if not verifier_cfg.get("id"):
                verifier_cfg["id"] = verifier_key
            if limit_agents is None or verifier_cfg["id"] in limit_agents:
                verifiers.append(verifier_cfg)
                verifier_ids.append(verifier_cfg["id"])
        if verifiers:
            for verifier_cfg in verifiers:
                await self._register_verifier(client, verifier_cfg)
        elif config_verifiers:
            LOGGER.info("No defined verifiers referenced by AGENTS")

    async def _register_verifier(self, client: IndyClient, verifier_cfg: dict) -> str:
        """
        Register a single verifier service from the configuration
        """
        verifier_id = verifier_cfg["id"]
        if "wallet" not in verifier_cfg:
            raise IndyConfigError("Wallet not defined for verifier: {}".format(verifier_id))
        wallet_cfg = verifier_cfg["wallet"]
        del verifier_cfg["wallet"]
        if "connection" not in verifier_cfg:
            raise IndyConfigError("Missing connection for verifier: {}".format(verifier_id))
        connection_cfg = verifier_cfg["connection"]
        del verifier_cfg["connection"]

        if not wallet_cfg.get("name"):
            wallet_cfg["name"] = verifier_id + "-Verifier-Wallet"
        if not wallet_cfg.get("seed"):
            raise IndyConfigError("Missing wallet seed for verifier: {}".format(verifier_id))
        wallet_id = await client.register_wallet(wallet_cfg)
        verifier_id = await client.register_verifier(wallet_id, verifier_cfg)

        if connection_cfg:
            if not connection_cfg.get("id"):
                connection_cfg["id"] = verifier_id
            conn_type = connection_cfg.get("type", "TheOrgBook")
            if conn_type == "TheOrgBook":
                _conn_id = await client.register_orgbook_connection(
                    verifier_id, connection_cfg)
            else:
                _conn_id = await client.register_http_connection(
                    verifier_id, connection_cfg)
        return verifier_id

    async def _register_proof_requests(self, client: IndyClient):
        """
        Register all proof request specifications from the configuration
        """
        config_prs = self.services_config("proof_requests")
        if config_prs:
            for pr_id, pr_spec in config_prs.items():
                if not pr_spec.get("id"):
                    pr_spec["id"] = pr_id
                _spec_id = await client.register_proof_spec(pr_spec)



"""
Helper methods provided for issue-credential credential generation
"""

import time
import datetime
from uuid import uuid4
from random import randrange


def uuid() -> str:
    """
    Generate a unique ID
    """
    return str(uuid4())


def pst():
    """
    Generate a random PST number
    """
    return '%d%d%d%d%d %d%d%d%d' % tuple(randrange(10) for i in range(9))


def now():
    """
    Get the current time as the number of seconds since the epoch
    """
    return int(time.mktime(
        datetime.datetime.now().timetuple()
    ))

def now_iso():
    """
    Get the current time, formatted as ISO-8601 string
    """
    return datetime.datetime.now().isoformat()


def one_year():
    """
    Get the time for a year from now, as the number of seconds since the epoch
    """
    return int(time.mktime(
        (datetime.datetime.now() + datetime.timedelta(days=365)).timetuple()
    ))

def one_year_iso():
    """
    Get the time for a year from now, formatted as ISO-8601 string
    """
    return (
          datetime.datetime.now() +
          datetime.timedelta(days=365)
        ).isoformat()



#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#pylint: disable=broad-except

"""
View classes for handling AJAX requests as an issuer or holder service
"""

import json
import logging

from aiohttp import web

from ..common.util import log_json, normalize_credential_ids
from ..indy.client import IndyClientError

from .view_helpers import (
    IndyRequestError,
    get_handle_id,
    get_manager,
    get_request_json,
    indy_client,
    perform_issue_credential,
    perform_store_credential,
    service_request,
)

LOGGER = logging.getLogger(__name__)


async def health(request: web.Request) -> web.Response:
    """
    Respond with HTTP code 200 if services are ready to accept new credentials, 451 otherwise
    """
    result = await get_manager(request).get_service_status('manager')
    ok = result and result.get("services", {}).get("indy", {}).get("synced")
    return web.Response(
        text='ok' if ok else '',
        status=200 if ok else 451)


async def status(request: web.Request) -> web.Response:
    """
    Respond with the current status of the application in JSON format
    """
    result = await get_manager(request).get_service_status('manager')
    return web.json_response(result)


async def ledger_status(request: web.Request) -> web.Response:
    """
    Respond with the status JSON retrieved from the Indy ledger (von-network)
    """
    #pylint: disable=broad-except
    result = await indy_client(request).get_ledger_status()
    try:
        jresult = json.loads(result)
        return web.json_response(jresult)
    except Exception:
        return web.Response(text=result)


async def hello(request: web.Request) -> web.Response:
    """
    Send a test request to the `HelloProcessor` service and return the response
    """
    result = await service_request(request, 'hello', 'isthereanybodyoutthere')
    return web.json_response(result)


async def issue_credential(request: web.Request, connection_id: str = None) -> web.Response:
    """
    Ask the Indy service to issue a credential to the Connection
    """
    try:
        connection_id = get_handle_id(request, "connection_id", connection_id)
        client = indy_client(request)
        params = await get_request_json(request)
        schema_name = request.query.get("schema")
        schema_version = request.query.get("version")
        stored, ret = await perform_issue_credential(
            client, connection_id, params, schema_name, schema_version)
    except IndyRequestError as e:
        return e.response

    response = web.json_response(ret)
    response["stored"] = stored
    return response


async def request_proof(request: web.Request, connection_id: str = None) -> web.Response:
    """
    Ask the Indy service to fetch a proof from the Connection
    """
    try:
        connection_id = get_handle_id(request, "connection_id", connection_id)
        inputs = await get_request_json(request)
        proof_name = request.query.get("name")
        if not proof_name:
            raise IndyRequestError("Missing 'name' parameter")
        params = {}
        cred_ids = request.query.get("credential_ids", request.query.get("credential_id"))
        if isinstance(inputs, dict):
            params = inputs.get("params", params)
            if not isinstance(params, dict):
                raise IndyRequestError("Parameter 'params' must be an object")
            cred_ids = normalize_credential_ids(inputs.get("credential_ids", cred_ids))
    except IndyRequestError as e:
        return e.response

    try:
        client = indy_client(request)
        proof_req = await client.generate_proof_request(proof_name)
        verified = await client.request_proof(connection_id, proof_req, cred_ids, params)
        result = {
            "verified": verified.verified,
            "parsed_proof": verified.parsed_proof,
            "proof": verified.proof.proof,
        }
        ret = {"success": True, "result": result}
    except IndyClientError as e:
        ret = {"success": False, "result": str(e)}
    return web.json_response(ret)


async def generate_credential_request(request, holder_id: str = None):
    """
    Processes a credential definition and responds with a credential request
    which can then be used to submit a credential.

    Example request payload:
    ```json
    {
        'credential_offer': <credential offer json>,
        'credential_definition_id': <credential definition id>
    }
    ```

    Returns:
    ```
    {
        "credential_request": <credential request json>,
        "credential_request_metadata": <credential request metadata json>
    }
    ```
    """

    try:
        holder_id = get_handle_id(request, "holder_id", holder_id)
        params = await get_request_json(request)
        offer = params.get("credential_offer")
        if not offer:
            raise IndyRequestError("Missing 'credential_offer'")
        cred_def = params.get("credential_definition")
        if cred_def:
            cred_def_id = cred_def.get("id")
        else:
            cred_def_id = params.get("credential_definition_id")
        if not cred_def_id:
            raise IndyRequestError("Missing 'credential_definition_id'")
    except IndyRequestError as e:
        return e.response

    try:
        cred_request = await indy_client(request).create_credential_request(
            holder_id, offer, cred_def_id)
        ret = {
            "success": True,
            "result": {
                "credential_request": cred_request.data,
                "credential_request_metadata": cred_request.metadata,
            }}
    except IndyClientError as e:
        cred_request = None
        ret = {"success": False, "result": str(e)}
    response = web.json_response(ret)
    response["cred_request"] = cred_request
    return response


async def store_credential(request, holder_id: str = None):
    """
    Stores a verifiable credential in wallet.
    The data in the credential is parsed and stored in the database
    for search/display purposes based on the issuer's processor config.
    The data is then made available through a REST API as well as a
    search API.

    Example request payload:
    ```json
    {
        "credential_data": <credential data>,
        "credential_request_metadata": <credential request metadata>
    }
    ```

    Returns: created verified credential model
    """
    try:
        holder_id = get_handle_id(request, "holder_id", holder_id)
        client = indy_client(request)
        params = await get_request_json(request)
        stored, ret = await perform_store_credential(client, holder_id, params)
    except IndyRequestError as e:
        return e.response

    response = web.json_response(ret)
    response["stored"] = stored
    return response


async def construct_proof(request, holder_id: str = None):
    """
    Constructs a proof given a proof request and source_id
   ```json
    {
        "proof_request": <HL Indy proof request>,
        "source_id": <source if of subject>
    }
    ```

    Returns: HL Indy proof data
    """
    try:
        holder_id = get_handle_id(request, "holder_id", holder_id)
        params = await get_request_json(request)
        proof_request = params.get("proof_request")
        if not proof_request:
            raise IndyRequestError("Missing 'proof_request'")
    except IndyRequestError as e:
        return e.response

    wql_filters = None # params.get("wql_filters")
    cred_ids = normalize_credential_ids(params.get("credential_ids"))
    try:
        LOGGER.debug("Performing proof request with cred IDs: %s", cred_ids)
        proof = await indy_client(request).construct_proof(
            holder_id, proof_request, wql_filters, cred_ids)
        ret = {"success": True, "result": proof.proof}
    except IndyClientError as e:
        proof = None
        ret = {"success": False, "result": str(e)}
    log_json("Proof response:", ret, LOGGER)
    response = web.json_response(ret)
    response["proof"] = proof
    return response



import os
import runpy
from setuptools import setup, find_packages

pkg_name = 'vonx'
meta = runpy.run_path("./{}/version.py".format(pkg_name))
version = meta['__version__']

setup(
    name=pkg_name,
    packages=find_packages(),
    package_data={'': ['*.html', '*.yml']},
    include_package_data=True,
    version=version,
    description='VON-X Connector',
    license='Apache Software License',
    author='PSPC-SPAC',
    author_email='andrew@1crm.com',
    url='https://github.com/PSPC-SPAC-buyandsell/von-x/',
    download_url='https://github.com/PSPC-SPAC-buyandsell/von-x/archive/v{}.tar.gz'.format(version),
    keywords=['verified-organizations-network', 'VON', 'TheOrgBook', 'Hyperledger', 'Indy', 'HTTP'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    python_requires='>=3.5.3',
    install_requires=[
        'aiohttp~=3.3.0',
        'aiohttp-jinja2~=1.1.0',
        'PyYAML',
        'von-anchor==1.6.36',
    ],
)



#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Common exception classes for Indy services
"""


class IndyError(Exception):
    """
    Base class for all exceptions thrown by :class:`IndyService` and related classes
    """
    pass

class IndyClientError(IndyError):
    """
    Base class for :class:`IndyClient`-related errors
    """
    pass

class IndyConfigError(IndyError):
    """
    Base class for :class:`IndyService` errors arising from configuration issues
    """
    pass

class IndyConnectionError(IndyError):
    """
    A generic exception representing an issue with a :class:`ConnectionBase` operation
    """

    def __init__(self, message: str, status=None, response=None):
        super(IndyConnectionError, self).__init__(message)
        self.message = message
        self.status = status
        self.response = response



import asyncio
from collections import OrderedDict
from concurrent.futures import Future
import json
import logging

from aiohttp import web

from ..common.exchange import RequestTarget
from ..indy.client import IndyClient, IndyClientError
from ..indy.errors import IndyError
from ..indy.messages import Credential, StoredCredential
from ..indy.manager import IndyManager

from .headers import KeyFinderBase, verify_signature

LOGGER = logging.getLogger(__name__)


class IndyRequestError(IndyError):
    """
    An exception in parsing request parameters
    """
    def __init__(self, message: str, *, status=400):
        super(IndyRequestError, self).__init__(message)
        self.message = message
        self.status = status

    @property
    def response(self):
        return web.Response(text=self.message, status=self.status)

class IndyCredentialProcessorException(IndyRequestError):
    """
    Base exception for :class:`IndyCredentialProcessor`
    """
    pass

class IndyCredentialProcessor:
    """
    Base class for post-storage credential processing operations
    """
    def __init__(self):
        pass

    def start_batch(self) -> object:
        """
        May return batch info used for caching and/or scheduling
        """
        pass

    def process_credential(
            self, stored: StoredCredential, origin_did: str = None, batch_info=None) -> Future:
        """
        Perform post-processing (DB indexing, etc)
        """
        return None

    async def process_credential_async(
            self, stored: StoredCredential, origin_did: str = None,
            batch_info=None) -> asyncio.Future:
        fut = self.process_credential(stored, origin_did, batch_info)
        if fut:
            return await asyncio.wrap_future(fut)

    def end_batch(self, batch_info):
        """
        Ensure that processing has been kicked off
        """
        pass


def get_manager(request: web.Request) -> IndyManager:
    """
    Fetch the service manager for the current application
    """
    return request.app['manager']

def get_request_target(request: web.Request, service_name: str) -> RequestTarget:
    """
    Create a :class:`RequestTarget` to process requests to a specific service

    Args:
        request: the incoming HTTP request
        service_name: the name of the service registered with the service manager
    """
    return get_manager(request).get_service_request_target(service_name)

def service_request(request: web.Request, service_name: str, message) -> Future:
    """
    Handle a single request to a running service and await the result in a thread

    Args:
        request: the incoming HTTP request
        service_name: the name of the service registered with the service manager
        message: the body of the message to be sent
    """
    return get_request_target(request, service_name).request(message)

async def get_request_json(request):
    try:
        return await request.json()
    except json.JSONDecodeError:
        raise IndyRequestError("Request body must contain an application/json payload")

def indy_client(request: web.Request) -> IndyClient:
    """
    Create an Indy client to perform requests against the ledger service
    """
    return get_manager(request).get_client()

def get_handle_id(request: web.Request, handle: str, override_val: str = None) -> str:
    """
    Check the request for a handle ID (connection or holder ID depending on the request)
    which may be overridden depending on the path
    """
    query_val = request.query.get(handle)
    match_val = override_val or request.match_info.get(handle)
    if query_val:
        if match_val and match_val != query_val:
            raise IndyRequestError("{} must be unspecified or equal to '{}'".format(handle, match_val))
    else:
        if not match_val:
            raise IndyRequestError("{} must be specified".format(handle))
        query_val = match_val
    return query_val

async def check_request_signature(
        request: web.Request, key_finder: KeyFinderBase, required: bool = False):
    """
    Check the DID-auth signature on the incoming request
    """
    if request.get("didauth"):
        return True, request["didauth"]
    auth = None
    try:
        auth = await verify_signature(
            request.headers, key_finder, request.method, request.path_qs)
    except IndyError as e:
        LOGGER.exception("Signature validation error:")
        raise IndyRequestError("Signature validation error") from e
    except Exception as e:
        LOGGER.exception("Signature validation error:")
        raise IndyRequestError("Not available", status=503) from e
    request["didauth"] = auth
    if not auth and required:
        raise IndyRequestError("Signature required", status=401)
    return auth

def get_request_did(request):
    auth = request.get("didauth")
    did = auth and auth["keyId"] or None
    if did and did.startswith("did:sov:"):
        did = did[8:]
    return did

def _assemble_cred_from_input(params: dict):
    """
    Assemble a single credential object from input
    """
    data = params.get("credential_data")
    if not data:
        raise IndyRequestError("Missing 'credential_data'")
    metadata = params.get("credential_request_metadata")
    if not metadata:
        raise IndyRequestError("Missing 'credential_request_metadata'")
    revoc_id = params.get("credential_revocation_id")
    return Credential(
        data,
        metadata,
        revoc_id,
    )


async def _issue_credential(
        client: IndyClient, connection_id: str,
        schema_name, schema_version, attribs, batch: bool = False):
    """
    Issue a single credential or batch of credentials
    """
    try:
        if batch:
            batch = await client.issue_credential_batch(
                connection_id, schema_name, schema_version, None, attribs)
            stored = []
            result = []
            erridx = 0
            for stored_cred in batch.results:
                if stored_cred.cred_id:
                    row = {"success": True, "result": stored_cred.cred_id}
                else:
                    errmsg = batch.errors[erridx] \
                        if batch.errors and erridx < len(batch.errors) else None
                    erridx += 1
                    row = {"success": False, "result": errmsg}
                if stored_cred.served_by:
                    row["served_by"] = stored_cred.served_by
                stored.append(stored_cred)
                result.append(row)
        else:
            stored = await client.issue_credential(
                connection_id, schema_name, schema_version, None, attribs)
            result = {"success": True, "result": stored.cred_id}
            if stored.served_by:
                result["served_by"] = stored.served_by
    except IndyClientError as e:
        stored = None
        result = {"success": False, "result": str(e)}
        if batch:
            stored = [stored for _row in attribs]
            result = [result for _row in attribs]
    return stored, result


async def perform_issue_credential(
        client: IndyClient, connection_id: str, params, schema_name=None, schema_version=None):
    """
    Parse request body into credential details and perform issuing
    """
    if isinstance(params, list):
        queue = OrderedDict()
        processed = {}
        orig_pos = []
        for cred in params:
            if not isinstance(cred, dict):
                raise IndyRequestError("Expected JSON object")
            if "schema" not in cred:
                raise IndyRequestError("Missing 'schema' property")
            if not isinstance(cred.get("attributes"), dict):
                raise IndyRequestError("Missing or non-dictionary 'attributes' property")
            key = (cred["schema"], cred.get("version"))
            if key not in queue:
                queue[key] = []
            orig_pos.append( (key, len(queue[key])) )
            queue[key].append(cred["attributes"])
        for key, attribs in queue.items():
            processed[key] = await _issue_credential(
                client, connection_id, key[0], key[1], attribs, True)
        stored = []
        result = []
        for key, pos in orig_pos:
            row = processed[key]
            stored.append(row[0][pos])
            result.append(row[1][pos])
        return stored, result
    else:
        if not schema_name:
            raise IndyRequestError("Missing 'schema' parameter")
        elif not isinstance(params, dict):
            raise IndyRequestError(
                "Request body must contain the credential attributes as a JSON object")
        return await _issue_credential(
            client, connection_id, schema_name, schema_version, params)


async def _store_credential(
        client: IndyClient, holder_id: str, cred: Credential,
        processor: IndyCredentialProcessor = None, origin_did: str = None,
        batch_info=None):
    """
    Process credential storage
    """
    stored = None
    try:
        stored = await client.store_credential(holder_id, cred)
        result = {"success": True, "result": stored.cred_id}
        if processor:
            await processor.process_credential_async(stored, origin_did, batch_info)
    except IndyClientError as e: # includes IndyCredentialProcessorException
        result = {"success": False, "result": str(e)}
    return stored, result


async def perform_store_credential(
        client: IndyClient, holder_id: str, params,
        processor: IndyCredentialProcessor = None, origin_did: str = None):
    """
    Parse request body into credential details and perform storage
    """

    if isinstance(params, list) and params:
        procs = []
        batch_info = processor and processor.start_batch() or None
        for cred in map(_assemble_cred_from_input, params):
            procs.append(asyncio.ensure_future(
                _store_credential(client, holder_id, cred, processor, origin_did, batch_info)))
        if batch_info:
            processor.end_batch(batch_info)
        await asyncio.wait(procs)
        stored = []
        result = []
        for proc in procs:
            stored_row, ret_row = proc.result()
            stored.append(stored_row)
            result.append(ret_row)
    elif isinstance(params, dict):
        cred = _assemble_cred_from_input(params)
        stored, result = await _store_credential(client, holder_id, cred, processor, origin_did)
    else:
        raise IndyRequestError(
            "Request body must contain the request parameters as a JSON object or list")
    return stored, result



#
# Copyright 2017-2018 Government of Canada
# Public Services and Procurement Canada - buyandsell.gc.ca
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Base implemention of Connections, to be managed by :class:`ConnectionCfg`.
"""

import asyncio
from enum import Enum
import logging
from typing import Sequence

import aiohttp

from ..common.exchange import RequestTarget
from .errors import IndyConfigError, IndyConnectionError
from .messages import (
    IndyServiceFail,
    CredentialOffer,
    Credential,
    GenerateCredentialRequestReq,
    CredentialRequest,
    StoreCredentialReq,
    StoredCredential,
    StoredCredentialBatch,
    ProofRequest,
    ConstructProofReq,
    ConstructedProof,
)

LOGGER = logging.getLogger(__name__)


class ConnectionType(Enum):
    """
    Enumeration of supported connection types
    """
    TheOrgBook = "TheOrgBook"
    holder = "holder"
    HTTP = "HTTP"


class HttpSession:
    """
    Handle an exception or bad response from an HTTP request
    """

    def __init__(self, method: str, http_client: aiohttp.ClientSession = None, timeout=None):
        self._client = http_client
        self._method = method
        self._opened = False
        self._timeout = timeout

    @property
    def client(self) -> aiohttp.ClientSession:
        """
        Accessor for the :class:`ClientSession` instance
        """
        return self._client

    async def check_status(self, response: aiohttp.ClientResponse, accept=(200, 201)):
        """
        Check the HTTP status of a response to a previous request
        """
        if response.status not in accept:
            raise IndyConnectionError(
                "Bad response from {}: ({}) {}".format(
                    self._method, response.status, await response.text()
                ),
                response.status,
                response,
            )

    async def __aenter__(self) -> 'ErrorHandler':
        if not self._client:
            self._client = aiohttp.ClientSession(read_timeout=self._timeout)
            self._opened = True
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        if self._opened:
            await self._client.close()
        if exc_type == IndyConnectionError:
            return False
        if exc_type == asyncio.TimeoutError:
            raise IndyConnectionError(
                "Connection timed out during {}".format(self._method),
                status=598,
            ) from None
        if exc_value:
            code = getattr(exc_value, 'code', None)
            raise IndyConnectionError(
                "Exception during {}: {} ({})".format(self._method, str(exc_value), code),
                status=code,
            ) from None


class ConnectionBase:
    """
    Base interface for Connection implementations
    """

    def __init__(self, agent_id: str, agent_type: str, agent_params: dict, conn_params: dict):
        self.agent_id = agent_id
        self.agent_type = agent_type
        self.agent_params = agent_params
        self.conn_params = conn_params
        self.created = False
        self.opened = False
        self.synced = False

    async def open(self, service: 'IndyService') -> None:
        """
        Initialize the connection
        """
        pass

    async def sync(self) -> None:
        """
        Perform any required synchronization
        """
        pass

    async def generate_credential_request(
            self, indy_offer: CredentialOffer) -> CredentialRequest:
        """
        Ask the target to generate a credential request from our credential offer

        Args:
            indy_offer: the result of preparing a credential offer
            creddef_id: the ID of the credential definition
        """
        pass

    async def store_credential(
            self, indy_cred: Credential) -> StoredCredential:
        """
        Ask the target to store a credential

        Args:
            indy_cred: the result of preparing a credential from a credential request
        """
        pass

    async def store_credential_batch(
            self, indy_creds: Sequence[Credential]) -> StoredCredentialBatch:
        """
        Ask the target to store a credential

        Args:
            indy_creds: the prepared list of credentials
        """
        results = []
        errors = []
        for cred in indy_creds:
            try:
                row = await self.store_credential(cred)
                if not row:
                    raise IndyConnectionError("not implemented")
                results.append(row)
            except IndyConnectionError as e:
                row = StoredCredential(
                    cred, None,
                )
                errors.append(str(e))
        return StoredCredentialBatch(results, errors)

    async def construct_proof(self, request: ProofRequest,
                              cred_ids: set = None, params: dict = None) -> ConstructedProof:
        """
        Ask the target to construct a proof from a proof request

        Args:
            request: the prepared Indy proof request
            params: extra parameters for the API
        """
        pass

    async def close(self) -> None:
        """
        Shut down the connection
        """
        pass


class HolderConnection(ConnectionBase):
    """
    :class:`ConnectionBase` interface implementation for local holder services
    """

    def __init__(self, agent_id: str, agent_type: str, agent_params: dict, conn_params: dict):
        super(HolderConnection, self).__init__(agent_id, agent_type, agent_params, conn_params)
        self.holder_id = self.conn_params.get("holder_id")
        if not self.holder_id:
            raise IndyConfigError("Missing 'holder_id' for holder connection")
        self.target = None

    async def open(self, service: 'IndyService') -> None:
        """
        Initialize the connection
        """
        self.target = RequestTarget(service, service.pid)

    async def close(self) -> None:
        """
        Shut down the connection
        """
        self.target = None

    async def generate_credential_request(
            self, indy_offer: CredentialOffer) -> CredentialRequest:
        """
        Ask the target to generate a credential request from our credential offer

        Args:
            indy_offer: the result of preparing a credential offer
        """
        result = await self.target.request(
            GenerateCredentialRequestReq(self.holder_id, indy_offer))
        if isinstance(result, IndyServiceFail):
            raise IndyConnectionError(result.value, 500)
        elif not isinstance(result, CredentialRequest):
            raise IndyConnectionError("Unexpected result: {}".format(result), 500)
        return result

    async def store_credential(
            self, indy_cred: Credential) -> StoredCredential:
        """
        Ask the target to store a credential

        Args:
            indy_cred: the result of preparing a credential from a credential request
        """
        result = await self.target.request(
            StoreCredentialReq(self.holder_id, indy_cred))
        if isinstance(result, IndyServiceFail):
            raise IndyConnectionError(result.value, 500)
        elif not isinstance(result, StoredCredential):
            raise IndyConnectionError("Unexpected result: {}".format(result), 500)
        return result

    async def construct_proof(self, request: ProofRequest,
                              cred_ids: set = None, params: dict = None) -> ConstructedProof:
        """
        Ask the target to construct a proof from a proof request

        Args:
            request: the prepared Indy proof request
            params: extra parameters for the API
        """
        result = await self.target.request(
            ConstructProofReq(self.holder_id, request, cred_ids))
        if isinstance(result, IndyServiceFail):
            raise IndyConnectionError(result.value, 500)
        elif not isinstance(result, ConstructedProof):
            raise IndyConnectionError("Unexpected result: {}".format(result), 500)
        return result


class HttpConnection(ConnectionBase):
    """
    A class for managing communication with an external agent over an HTTP connection
    """

    def __init__(self, agent_id: str, agent_type: str, agent_params: dict, conn_params: dict):
        super(HttpConnection, self).__init__(agent_id, agent_type, agent_params, conn_params)
        self._api_url = self.conn_params.get("api_url")
        if not self._api_url:
            raise IndyConfigError("Missing 'api_url' for HTTP connection")
        self._http_client = None
        self._response_headers = None

    async def open(self, service: "IndyService") -> None:
        # TODO check DID is registered etc ..
        self._http_client = service._connection_http_client(self.conn_params["id"])

    async def close(self) -> None:
        """
        Shut down the connection
        """
        if self._http_client:
            await self._http_client.close()
            self._http_client = None

    @property
    def path_prefix(self):
        return ''

    async def generate_credential_request(
            self, indy_offer: CredentialOffer) -> CredentialRequest:
        """
        Ask the API to generate a credential request from our credential offer

        Args:
            indy_offer: the result of preparing a credential offer
        """
        response = await self.post_json(
            self.path_prefix + "generate-credential-request", {
                "credential_offer": indy_offer.data,
                "credential_definition_id": indy_offer.cred_def_id,
            }
        )
        LOGGER.debug("Credential request response: %s", response)
        result = response.get("result")
        if not response.get("success"):
            raise IndyConnectionError(
                "Could not create credential request: {}".format(result),
                400,
                response,
            )
        return CredentialRequest(
            indy_offer,
            result["credential_request"],
            result["credential_request_metadata"],
        )

    async def store_credential(
            self, indy_cred: Credential) -> StoredCredential:
        """
        Ask the API to store a credential

        Args:
            indy_cred: the result of preparing a credential from a credential request
        """
        response = await self.post_json(
            self.path_prefix + "store-credential", {
                "credential_data": indy_cred.cred_data,
                "credential_request_metadata": indy_cred.cred_req_metadata,
            }
        )
        LOGGER.debug("Store credential response: %s", response)
        result = response.get("result")
        if not response.get("success"):
            raise IndyConnectionError(
                "Credential was not stored: {}".format(result),
                400,
                response,
            )
        served_by = self._response_headers and self._response_headers.get('X-Served-By')
        return StoredCredential(
            indy_cred,
            result,
            served_by,
        )

    async def store_credential_batch(
            self, indy_creds: Sequence[Credential]) -> StoredCredentialBatch:
        """
        Ask the API to store a list of credentials

        Args:
            indy_creds: the prepared list of credentials
        """
        data = [
            {
                "credential_data": cred.cred_data,
                "credential_request_metadata": cred.cred_req_metadata,
            } for cred in indy_creds
        ]
        response = await self.post_json(self.path_prefix + "store-credential", data)
        LOGGER.debug("Store credential batch response: %s", response)

        served_by = self._response_headers and self._response_headers.get('X-Served-By')
        stored = []
        errors = []
        for idx, row in enumerate(response):
            result = row.get("result")
            success = row.get("success")
            stored.append(StoredCredential(
                indy_creds[idx],
                success and result or None,
                served_by,
            ))
            if not success:
                errors.append("Credential was not stored: {}".format(result))
        return StoredCredentialBatch(
            stored, errors,
        )

    async def construct_proof(self, request: ProofRequest,
                              cred_ids: set = None, params: dict = None) -> ConstructedProof:
        """
        Ask the API to construct a proof from a proof request

        Args:
            proof_request: the prepared Indy proof request
        """
        response = await self.post_json(
            self.path_prefix + "construct-proof", {
                "source_id": params and params.get("source_id") or None,
                "proof_request": request.data,
                "credential_ids": list(cred_ids) if cred_ids else None,
            }
        )
        result = response.get("result")
        if not response.get("success"):
            raise IndyConnectionError(
                "Error constructing proof: {}".format(result),
                400,
                response,
            )
        return ConstructedProof(
            result,
        )

    def get_api_url(self, path: str = None) -> str:
        """
        Construct the URL for an API request

        Args:
            path: an optional path to be appended to the URL
        """
        url = self._api_url
        if not url.endswith("/"):
            url += "/"
        if path:
            url = url + path
        return url

    async def post_json(self, path: str, data):
        """
        A standard POST request to an API method

        Args:
            path: The relative path to the API method
            data: The body of the request, to be converted to JSON

        Returns:
            the decoded JSON response
        """
        url = self.get_api_url(path)
        LOGGER.debug("post_json: %s", url)
        async with HttpSession("post_json", self._http_client) as handler:
            response = await handler.client.post(url, json=data)
            self._response_headers = response.headers
            await handler.check_status(response)
            return await response.json()



import asyncio
import logging
import os
import sys

from vonx.indy.manager import IndyManager

LOGGER = logging.getLogger()


class TestIndyHolder(IndyManager):
    """
    A test Indy service manager which creates a sample holder
    """

    def get_service_init_params(self) -> dict:
        return {
            "auto_register": 1,
            "genesis_path": "/home/indy/genesis",
            "ledger_url": os.environ.get("LEDGER_URL") or "http://192.168.65.3:9000",
        }

    async def _load_config(self) -> None:
        # skip default implementation
        pass

    async def add_test_services(self, client):
        LOGGER.info("setting up test indy holder")

        all = {}
        all["holder_wallet_id"] = await client.register_wallet({
            "name": "holder-wallet",
            "seed": "holder-wallet-000000000000000002",
        })
        all["holder_id"] = await client.register_holder(all["holder_wallet_id"], {
            "id": "holder",
            "name": "Test Holder",
        })

        return all


def test_web(manager):
    from vonx.web import init_web
    app = init_web(manager)

    from aiohttp import web
    web.run_app(app, host='0.0.0.0', port=5000)


if __name__ == '__main__':
    LOGGER.setLevel(logging.DEBUG)
    CONSOLE = logging.StreamHandler()
    CONSOLE.setLevel(os.environ.get("LOG_LEVEL", logging.INFO))
    LOGGER.addHandler(CONSOLE)

    MGR = TestIndyHolder()
    MGR.start()

    CLIENT = MGR.get_client()
    DONE = False
    async def setup(client, teardown=False):
        ids = await MGR.add_test_services(client)
        if teardown:
            MGR.stop()

    try:
        asyncio.get_event_loop().run_until_complete(setup(CLIENT))
        test_web(MGR)
    except:
        LOGGER.exception("Error during init")
        MGR.stop()

    LOGGER.info("done")
